#!/usr/bin/env python3

#	MetaCall Python Port by Parra Studios
#	A frontend for Python language bindings in MetaCall.
#
#	Copyright (C) 2016 - 2025 Vicente Eduardo Ferrer Garcia <vic798@gmail.com>
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#		http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

from metacall.api import metacall, metacall_load_from_file, metacall_load_from_memory, metacall_load_from_package, metacall_inspect, metacall_value_create_ptr, metacall_value_reference, metacall_value_dereference



#!/usr/bin/env python3

#	MetaCall Python Port by Parra Studios
#	A frontend for Python language bindings in MetaCall.
#
#	Copyright (C) 2016 - 2025 Vicente Eduardo Ferrer Garcia <vic798@gmail.com>
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#		http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

import os
import re
import sys
import json
import ctypes

def find_files_recursively(root_dir, pattern):
	regex = re.compile(pattern)
	matches = []
	for dirpath, dirnames, filenames in os.walk(root_dir):
		for filename in filenames:
			if regex.search(filename):
				matches.append(os.path.join(dirpath, filename))
	return matches

def platform_install_paths():
	if sys.platform == 'win32':
		return {
			'paths': [ os.path.join(os.environ.get('LOCALAPPDATA', ''), 'MetaCall', 'metacall') ],
			'name': r'metacall\.dll'
		}
	elif sys.platform == 'darwin':
		return {
			'paths': [ '/opt/homebrew/lib/', '/usr/local/lib/' ],
			'name': r'libmetacall\.dylib'
		}
	elif sys.platform == 'linux':
		return {
			'paths': [ '/usr/local/lib/', '/gnu/lib/' ],
			'name': r'libmetacall\.so'
		}
	else:
		raise RuntimeError(f"Platform {sys.platform} not supported")

def search_paths():
	custom_path = os.environ.get('METACALL_INSTALL_PATH')
	if custom_path:
		return {
			'paths': [ custom_path ],
			'name': r'^(lib)?metacall(d)?\.(so|dylib|dll)$'
		}

	return platform_install_paths()

def find_library():
	search_data = search_paths()

	for path in search_data['paths']:
		files = find_files_recursively(path, search_data['name'])
		if files:
			return files[0]

	raise ImportError("""
		MetaCall library not found, if you have it in a special folder, define it through METACALL_INSTALL_PATH'.
		"""
		+ "Looking for it in the following paths: " + ', '.join(search_data['paths']) + """
		If you do not have it installed, you have three options:
			1) Go to https://github.com/metacall/install and install it.
			2) Contribute to https://github.com/metacall/distributable by providing support for your platform and architecture.
			3) Be a x10 programmer and compile it by yourself, then define the install folder if it is different from the default in os.environ['METACALL_INSTALL_PATH'].
	""")

def metacall_module_load():
	# Check if it is loaded from MetaCall or from Python
	if 'py_port_impl_module' in sys.modules:
		return sys.modules['py_port_impl_module']

	# Define the Python host
	os.environ['METACALL_HOST'] = 'py'

	# Find the shared library
	library_path = find_library()

	# Load MetaCall
	lib = ctypes.CDLL(library_path, mode=ctypes.RTLD_GLOBAL)

	# Python Port must have been loaded at this point
	if 'py_port_impl_module' in sys.modules:
		port = sys.modules['py_port_impl_module']

		# For some reason, Windows deadlocks on initializing asyncio
		# but if it is delayed, it works, so we initialize it after here
		if sys.platform == 'win32':
			port.py_loader_port_asyncio_initialize()

		return port
	else:
		raise ImportError(
			'MetaCall was found but failed to load'
		)

# Load metacall extension depending on the platform
module = metacall_module_load()

# Load from file
def metacall_load_from_file(tag, paths):
	return module.metacall_load_from_file(tag, paths)

# Load from package
def metacall_load_from_package(tag, path):
	return module.metacall_load_from_package(tag, path)

# Load from memory
def metacall_load_from_memory(tag, buffer):
	return module.metacall_load_from_memory(tag, buffer)

# Invocation
def metacall(function_name, *args):
	return module.metacall(function_name, *args)

# Wrap metacall inspect and transform the json string into a dict
def metacall_inspect():
	data = module.metacall_inspect()
	if data:
		dic = json.loads(data)
		try:
			del dic['__metacall_host__']
		except:
			pass
		return dic
	return dict()

# Value API for handling pointers
def metacall_value_create_ptr(ptr):
	return module.metacall_value_create_ptr(ptr)

# Value API for getting the pointer to a value
def metacall_value_reference(v):
	return module.metacall_value_reference(v)

# Value API for getting the value of a pointer
def metacall_value_dereference(ptr):
	return module.metacall_value_dereference(ptr)

# Monkey patching
import builtins
import types
from contextlib import suppress
import functools

# Save the original Python import
__python_import__ = builtins.__import__

# Define proper import exception depending on Python support
ImportException = ImportError

try:
	try:
		raise ModuleNotFoundError()
	except ModuleNotFoundError as e:
		ImportException = ModuleNotFoundError
except NameError as e:
	ImportException = ImportError

def __metacall_import__(name, globals=None, locals=None, fromlist=(), level=0):

	def find_handle(handle_name):
		metadata = metacall_inspect()

		for loader in metadata.keys():
			for handle in metadata[loader]:
				if handle['name'] == handle_name:
					return dict(functools.reduce(lambda symbols, func: {**symbols, func['name']: lambda *args: metacall(func['name'], *args) }, handle['scope']['funcs'], {}))

		return None

	def generate_module(handle_name, handle):
		mod = sys.modules.setdefault(handle_name, types.ModuleType(handle_name))

		# Set a few properties required by PEP 302
		base_path = os.environ.get('LOADER_SCRIPT_PATH', os.getcwd())
		mod.__file__ = os.path.join(base_path, handle_name)
		mod.__name__ = handle_name
		mod.__path__ = base_path
		# TODO: Using os.__loader__ instead of self until we implement the custom loader class
		mod.__loader__ = os.__loader__ # self
		# PEP-366 specifies that package's set __package__ to
		# their name, and modules have it set to their parent package (if any)
		# TODO (https://pymotw.com/3/sys/imports.html):
		# if self.is_package(handle_name):
		# 	mod.__package__ = handle_name
		# else:
		# 	mod.__package__ = '.'.join(handle_name.split('.')[:-1])
		mod.__package__ = handle_name

		# Add the symbols to the module
		mod.__dict__.update(handle)

		return mod

	# Map file extension to tags
	file_extensions_to_tag = {
		# Mock Loader
		'mock': 'mock',
		# Python Loader
		'py': 'py',
		# Ruby Loader
		'rb': 'rb',
		# C# Loader
		'cs': 'cs',
		'vb': 'cs',
		# Cobol Loader
		'cob': 'cob',
		'cbl': 'cob',
		'cpy': 'cob',
		# NodeJS Loader
		'node': 'node', # TODO: Load by package on node is not implemented or it is unnecesary
		'js': 'node',
		# WebAssembly Loader
		'wat': 'wasm',
		# TypeScript Loader
		'ts': 'ts',
		'jsx': 'ts',
		'tsx': 'ts',
		# Rust Loader
		'rs': 'rs',
		# C Loader
		'c': 'c'

		# Note: By default js extension uses NodeJS loader instead of JavaScript V8
		# Probably in the future we can differenciate between them, but it is not trivial
	}

	package_extensions_to_tag = {
		# C# Loader
		'dll': 'cs',
		# WebAssembly Loader
		'wasm': 'wasm',
		# Rust Loader
		'rlib': 'rs',
	}

	# Set containing all tags
	available_tags = set({**file_extensions_to_tag, **package_extensions_to_tag})

	# Try to load by module: import metacall.node.ramda
	metacall_module_import = name.split('.')
	metacall_module_import_size = len(metacall_module_import)

	if metacall_module_import_size > 1 and metacall_module_import[0] == 'metacall':
		if not metacall_module_import[1] in available_tags:
			raise ImportException(f'MetaCall could not import: {name}; the loader {metacall_module_import[1]} does not exist')

		def _metacall_module_import(module_tag, module_name, return_root):
			# Check if it is already loaded
			metacall_module = sys.modules['metacall']
			if hasattr(metacall_module, module_tag):
				metacall_module_tag = getattr(metacall_module, module_tag)
				if hasattr(metacall_module_tag, module_name):
					if return_root == True:
						return getattr(metacall_module_tag, module_name)
					else:
						return metacall_module_tag

			handle = module.metacall_load_from_file_export(
				module_tag, [module_name]
			)

			if handle != None:
				# Generate the module
				imported_module = generate_module(module_name, handle)

				if imported_module is None:
					raise ImportException(f'MetaCall could not import: {name}; failed to generate the module')

				# Add the new functions to metacall module
				metacall_module = sys.modules['metacall']
				if not hasattr(metacall_module, module_tag):
					setattr(metacall_module, module_tag, types.ModuleType(module_tag))

				metacall_module_tag = getattr(metacall_module, module_tag)
				setattr(metacall_module_tag, module_name, imported_module)
				setattr(metacall_module, module_tag, metacall_module_tag)

				if return_root == True:
					return metacall_module
				else:
					return metacall_module_tag
			else:
				raise ImportException(f'MetaCall could not import: {name}')

		# Case 1: import metacall.node.ramda && Case 2: from metacall.node.ramda import func1, func2, func3
		if metacall_module_import_size >= 3:
			module_tag = metacall_module_import[1]
			module_name = '.'.join(metacall_module_import[2:])
			return _metacall_module_import(module_tag, module_name, True)

		# Case 3: from metacall.node import ramda, express
		elif metacall_module_import_size == 2:
			if fromlist is None:
				raise ImportException(f'MetaCall could not import: {name}; you need to specify the submodules: from {name} import moduleA, moduleB')

			metacall_module_tag = None
			module_tag = metacall_module_import[1]

			for module_name in fromlist:
				metacall_module_tag = _metacall_module_import(module_tag, module_name, False)

			return metacall_module_tag

	# Try to load by extension: import script.js
	filename, extension = os.path.splitext(name)

	if extension != '':
		ext = extension[1:]

		if (
			ext in file_extensions_to_tag.keys()
			or ext in package_extensions_to_tag.keys()
		):
			# Check if it is already loaded in MetaCall
			handle = find_handle(name)

			if handle != None:
				# Generate the module from cached handle
				return generate_module(filename, handle)

			if ext in file_extensions_to_tag.keys():
				# Load from file
				file_tag = file_extensions_to_tag[ext]

				if file_tag == 'py':
					return __python_import__(filename, globals, locals, fromlist, level)
				else:
					handle = module.metacall_load_from_file_export(file_tag, [name])

			elif ext in package_extensions_to_tag.keys():
				# Load from package
				package_tag = package_extensions_to_tag[ext]

				if package_tag == 'py':
					return __python_import__(filename, globals, locals, fromlist, level)
				else:
					handle = module.metacall_load_from_package_export(package_tag, name)

			if handle != None:
				# Generate the module
				return generate_module(filename, handle)

	# Load with Python
	return __python_import__(name, globals, locals, fromlist, level)

# Override Python import
builtins.__import__ = __metacall_import__



#!/usr/bin/env python3

#	MetaCall Python Port by Parra Studios
#	A frontend for Python language bindings in MetaCall.
#
#	Copyright (C) 2016 - 2025 Vicente Eduardo Ferrer Garcia <vic798@gmail.com>
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#		http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

# To use a consistent encoding
from codecs import open
import os
import sys
import re

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

current_path = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(current_path, 'README.rst'), encoding='utf-8') as f:
	long_description = f.read()

# Define set up options
options = {
	'name': 'metacall',

	# Versions should comply with PEP440.  For a discussion on single-sourcing
	# the version across setup.py and the project code, see
	# https://packaging.python.org/en/latest/single_source_version.html
	'version': '0.5.3',

	'description': 'A library for providing inter-language foreign function interface calls',
	'long_description': long_description,
	'long_description_content_type': 'text/x-rst',

	# The project's main homepage
	'url': 'https://github.com/metacall/core',

	'project_urls': {
		'Homepage': 'https://metacall.io',
		'Repository': 'https://github.com/metacall/core',
		'Documentation': 'https://core.metacall.io/',
		'Issues': 'https://github.com/metacall/core/issues',
	},

	# Author details
	'author': 'Vicente Eduardo Ferrer Garcia',
	'author_email': 'vic798@gmail.com',

	# License
	'license': 'Apache License 2.0',

	# See https://pypi.python.org/pypi?%3Aaction=list_classifiers
	'classifiers': [
		# Project Status
		#   3 - Alpha
		#   4 - Beta
		#   5 - Production/Stable
		'Development Status :: 4 - Beta',

		# Audience
		'Intended Audience :: Developers',
		'Topic :: Software Development :: Interpreters',

		# Python versions support
		#'Programming Language :: Python :: 2',
		#'Programming Language :: Python :: 2.6',
		#'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Programming Language :: Python :: 3.10',
		'Programming Language :: Python :: 3.11',
		'Programming Language :: Python :: 3.12',
		'Programming Language :: Python :: 3.13',
	],

	# Keywords
	'keywords': 'metacall python port ffi polyglot faas serverless',

	# Required Python version
	'python_requires': '>=3.3',

	# Dependencies
	'install_requires': [],

	# List additional groups of dependencies here (e.g. development
	# dependencies). You can install these using the following syntax,
	# for example:
	# $ pip install -e .[dev,test]
	'extras_require': {
		'dev': ['check-manifest'],
		'test': ['coverage'],
	},

	# If there are data files included in your packages that need to be
	# installed, specify them here.  If using Python 2.6 or less, then these
	# have to be included in MANIFEST.in as well.
	#package_data: {
	#	'sample': ['package_data.dat'],
	#},

	# Although 'package_data' is the preferred approach, in some case you may
	# need to place data files outside of your packages. See:
	# http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
	# In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
	#data_files: [('my_data', ['data/data_file'])],
}

# Exclude base packages
exclude_packages = ['contrib', 'docs', 'test', 'test.py', 'CMakeLists.txt', '.gitignore', 'upload.sh']

# Define required packages
options['packages'] = find_packages(exclude=exclude_packages)

# Execute the setup
setup(**options)



#!/usr/bin/env python3

import os
import sys
import unittest

# Load metacall from Python Port path
abspath = os.path.dirname(os.path.abspath(__file__))
relpath = '..'
path = os.path.normpath(os.path.join(abspath, relpath))

# Insert first in the sys path so we make sure we load the correct port
sys.path.insert(0, path)

from metacall import *

class py_port_test(unittest.TestCase):

	# MetaCall (Mock)
	def test_mock(self):
		self.assertEqual(metacall_load_from_file('mock', ['sometestmock.mock']), True)

		self.assertEqual(metacall('three_str', 'a', 'b', 'c'), 'Hello World')

	# MetaCall (Python from memory)
	def test_python_memory(self):
		script = '#!/usr/bin/env python3\ndef inline_multiply_mem(left, right):\n\treturn left * right\n'

		self.assertEqual(metacall_load_from_memory('py', script), True)

		self.assertEqual(metacall('inline_multiply_mem', 2, 2), 4)

	# MetaCall (Python)
	def test_python(self):
		import example

		self.assertEqual(example.hello(), None)

		self.assertEqual(example.multiply(5, 7), 35)

		self.assertEqual(metacall_load_from_file('py', ['helloworld.py']), True)

		self.assertEqual(metacall('s_sum', 5, 5), 10)

	# MetaCall (Ruby)
	def test_ruby(self):
		from second.rb import get_second, get_second_untyped

		self.assertEqual(get_second(34, 22), 22)

		self.assertEqual(get_second_untyped(34, 22), 22)

		self.assertEqual(metacall_load_from_file('rb', ['hello.rb']), True)

		self.assertEqual(metacall('say_null'), None)

		self.assertEqual(metacall('say_multiply', 3, 4), 12)

		self.assertEqual(metacall('say_hello', 'world'), 'Hello world!')

	# MetaCall (NodeJS)
	def test_nodejs(self):
		from derpyramda.js import add, subtract, flip

		self.assertEqual(add(3, 4), 7.0)
		self.assertEqual(subtract(5, 4), 1.0)

		# TODO:
		# The current code does the following:
		#	flip is a Python function which wraps a NodeJS function,
		#	when calling to the function in NodeJS, the call gets enqueued
		#	in order to make a thread safe call into the V8 thread. Once this
		#	happens, flip calls back into a Python lambda (in the first case),
		#	or function (wrapping subtract) in the second case. As the Python GIL is
		#	already acquired by the first thread, the main thread which did the call,
		#	different from the V8 thread, the GIL gets aquired in the new thread.
		#
		#		|          Python Thread          |        V8 Thread       |
		#		|----------------------------------------------------------|
		#		| GIL lock                        |                        |
		#		|    Call flip                    |                        |
		#		|        Enqueue call in V8       |                        |
		#		|        Wait for V8              |                        |
		#		|                                 | Execute flip           |
		#		|                                 |    Call lambda         |
		#		|                                 |        GIL lock        |
		#
		# For avoiding extra work when executing calls in the same thread, this is the mechanism that we can use:
		#	1) Get the thread id of the thread where Python was launched, similarly to this:
		#		https://github.com/metacall/core/blob/9ad4ed8964a53e30d8ab478a53122c396d705cdd/source/loaders/node_loader/source/node_loader_impl.cpp#L3158
		#	2) Check if the current thread is the same as where Python interpreter was launched and check against this:
		#		PyGILState_Check: https://python.readthedocs.io/en/latest/c-api/init.html#c.PyGILState_Check
		#
		self.assertEqual(flip(lambda x, y: x - y)(5, 4), -1.0)
		self.assertEqual(flip(subtract)(5, 4), -1.0)

	# MetaCall (Rust)
	def test_rust(self):
		if os.environ.get('OPTION_BUILD_LOADERS_RS') == None: return

		from basic.rs import add, add_float, string_len, new_string

		self.assertEqual(add(34, 22), 56)

		self.assertEqual(add_float(34.0, 22.0), 56.0)

		# FIXME: python will get a string with random suffix
		# s = new_string(3)
		# self.assertEqual(s, 'get number 3')

		self.assertEqual(string_len('world'), 5)

		metacall_load_from_memory("rs", """
fn return_string() -> String {
	println!("return hello world");
    String::from("hello world")
}

fn new_string2(idx: i32) -> String {
    format!("get number {idx}")
}
""")
		self.assertEqual(metacall('new_string2', 5), 'get number 5')

		# FIXME: println inside rust fails to log anything
		self.assertEqual(metacall("return_string"), "hello world")

if __name__ == '__main__':
	unittest.main()
